/* 
 *   Copyright (c) 1997-2001 Michael Hoffmann (hoffmann@inf.ethz.ch)
 *   
 *   Permission is hereby granted, free of charge, to any person
 *   obtaining a copy of this software and associated documentation
 *   files (the "Software"), to deal in the Software without
 *   restriction, including without limitation the rights to use, copy,
 *   modify, merge, publish, distribute, sublicense, and/or sell copies
 *   of the Software, and to permit persons to whom the Software is
 *   furnished to do so, subject to the following conditions:
 *      
 *   The above copyright notice and this permission notice shall be
 *   included in all copies or substantial portions of the Software.
 *
 *   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 *   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 *   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 *   NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 *   BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 *   ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 *   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 *   SOFTWARE.
 *      
 *   $Revision: 1.5 $ 
 *   $Date: 2001/09/20 09:28:49 $
 *      
 *   convert C++ source code to highlighted latex (using colordvi) 
 */

%option noyywrap
%s MAY_KEY
%s MAY_CLASS_DECL
%s MAY_BASE_CLASS_NAME
%s MAY_TYPE_DECL
%s MAY_TYPE_NAME
%s MAY_USING
%s MAY_NAMESPACE

   /* C style comments are too nasty ;) */
COMMENT		"//".*
STRING		\"[^"\n]*\"

WHITE		[ \t]
WHITE_CR	{WHITE}|\n
WHITE_COMMENT	({WHITE_CR}|{COMMENT})*
ALPHA		[a-zA-Z_]
ALNUM 		{ALPHA}|[[:digit:]]

D1_PAREN_EXPR 	"{"[^{}]*"}"
D2_PAREN_EXPR 	"{"({D1_PAREN_EXPR}|[^{}])*"}"
D3_PAREN_EXPR	"{"({D2_PAREN_EXPR}|[^{}])*"}"
D4_PAREN_EXPR	"{"({D3_PAREN_EXPR}|[^{}])*"}"
D5_PAREN_EXPR	"{"({D4_PAREN_EXPR}|[^{}])*"}"
D6_PAREN_EXPR	"{"({D5_PAREN_EXPR}|[^{}])*"}"
PAREN_EXPR 	{D6_PAREN_EXPR}

D1_RPAREN_EXPR 	"("[^()]*")"
D2_RPAREN_EXPR 	"("({D1_RPAREN_EXPR}|[^()])*")"
D3_RPAREN_EXPR	"("({D2_RPAREN_EXPR}|[^()])*")"
D4_RPAREN_EXPR	"("({D3_RPAREN_EXPR}|[^()])*")"
D5_RPAREN_EXPR	"("({D4_RPAREN_EXPR}|[^()])*")"
D6_RPAREN_EXPR	"("({D5_RPAREN_EXPR}|[^()])*")"
RPAREN_EXPR 	{D6_RPAREN_EXPR}

D1_SPAREN_EXPR 	"<"[^<>()]*">"
D2_SPAREN_EXPR 	"<"({D1_SPAREN_EXPR}|[^<>()])*">"
D3_SPAREN_EXPR	"<"({D2_SPAREN_EXPR}|[^<>()])*">"
D4_SPAREN_EXPR	"<"({D3_SPAREN_EXPR}|[^<>()])*">"
D5_SPAREN_EXPR	"<"({D4_SPAREN_EXPR}|[^<>()])*">"
D6_SPAREN_EXPR	"<"({D5_SPAREN_EXPR}|[^<>()])*">"
SPAREN_EXPR 	{D6_SPAREN_EXPR}

BOL		{WHITE}*
BASE_ID		"~"?{ALPHA}{ALNUM}*
    /* SPARENEXPR for template parameters */
ID 		{BASE_ID}{SPAREN_EXPR}?
QUAL_ID 	"::"?{ID}("::"{ID})*([&*]|"[]")*
OPERATOR	operator{WHITE}*([-!%^&*+=~|<>,/]+|(new|delete)"[]"?|"[]"|"()")
    /* this is particularly nasty, but needed to avoid that
       operator is parsed as BASE_ID, too */
ALNUM_NO_O	[[:digit:]]|[a-np-zA-Z_]
ALNUM_NO_P	[[:digit:]]|[a-oq-zA-Z_]
ALNUM_NO_E	[[:digit:]]|[a-df-zA-Z_]
ALNUM_NO_R	[[:digit:]]|[a-qs-zA-Z_]
ALNUM_NO_A	[[:digit:]]|[b-zA-Z_]
ALNUM_NO_T	[[:digit:]]|[a-su-zA-Z_]
BASE_ID_NO_OP	"~"?([a-np-zA-Z_]|o{ALNUM_NO_P}|op{ALNUM_NO_E}|ope{ALNUM_NO_R}|oper{ALNUM_NO_A}|opera{ALNUM_NO_T}|operat{ALNUM_NO_O}|operato{ALNUM_NO_R}){ALNUM}*
FUNC_NAME 	({BASE_ID_NO_OP}|{OPERATOR}){SPAREN_EXPR}?
QUAL_FUNC_NAME	"::"?{FUNC_NAME}("::"{FUNC_NAME})*

    /* after these a keyword may follow */
PRECEDE_KEYWORD	{WHITE}|"("
    /* these have to follow a keyword */
FOLLOW_KEYWORD	[ (;\n\t&*:]

    /* important keywords */
IMP_KEYWORD	(asm|do|sizeof|break|else|case|catch|switch|this|for|throw|while|continue|goto|try|delete|if|return|new|exit|abort)

    /* access specifier */
ACCESS_SPEC	p(ublic|r(ivate|otected))

    /* keywords opening declarations */
DECL_KEYWORD1	typedef|typename|namespace|using

    /* keywords opening declarations of class type */
DECL_KEYWORD2	class|union|enum|struct

    /* builtin types */
TYPE_KEYWORD	short|double|int|bool|long|char|void|float|wchar_t|false|true|default

    /* cv qualifier */
CV_QUAL		volatile|const

    /* modifier keywords */
MOD_KEYWORD	inline|signed|mutable|static|unsigned|explicit|extern|register|auto|friend|virtual|{CV_QUAL}

    /* other keywords */
OTHER_KEYWORD	typeid|dynamic_cast|static_cast|const_cast|reinterpret_cast|assert

    /* preprocessor directives */
PREPROC1		"#"{WHITE}*(include|if|endif|else)[^\n]*
PREPROC2		"#"{WHITE}*(define|undef)[^\n]*

    /* constructor initialisation sequence
       (also used for inheritance decls) */
CONS_INIT		":"[^{;]*
CLASS_DECL_FOLLOW1	{WHITE_COMMENT}{PAREN_EXPR};
CLASS_DECL_FOLLOW2	{WHITE_COMMENT};
CLASS_DECL_FOLLOW3	{WHITE_COMMENT}{CONS_INIT}{PAREN_EXPR};
CLASS_DECL 		{WHITE}+{ID}{WHITE_COMMENT}{CONS_INIT}?{PAREN_EXPR}?;

NAMESPACE		namespace
NAMESPACE_FOLLOW 	{WHITE}*{PAREN_EXPR}
FULL_NAMESPACE_FOLLOW 	{WHITE}+{ID}{NAMESPACE_FOLLOW}

USING			using
FULL_USING_FOLLOW 	{WHITE}+({QUAL_ID}|{NAMESPACE}{WHITE}+{ID});

EXPR			({RPAREN_EXPR}|[^(){},])+
FULL_TYPE		{QUAL_ID}({WHITE}+{QUAL_ID})*
FUNC_ARG 		{FULL_TYPE}({WHITE}*={EXPR})?
FUNC_ARGLIST 		{FUNC_ARG}({WHITE}*","{WHITE_CR}*{FUNC_ARG})*
FUNC_DECL_FOLLOW 	{WHITE}*"("{WHITE_CR}*({FUNC_ARGLIST}{WHITE}*)?")"{WHITE_COMMENT}({CV_QUAL}{WHITE_COMMENT})*{CONS_INIT}?({PAREN_EXPR}|";")
TEMPLATE_DECL	"template"{WHITE_CR}*{SPAREN_EXPR}

%{
#define RED	"\\Mahogany{" 
#define BLUE	"\\Orchid{" 
#define LBLUE	"\\MidnightBlue{" 
#define MBLUE	"\\Periwinkle{" 
#define GBLUE	"\\CadetBlue{" 
#define BLACK	"\\Black{" 
#define TAN	"\\Tan{" 
#define VIOLET	"\\Blue{" 
#define MAGENTA	"\\Thistle{" 
#define BROWN	"\\Plum{" 
#define GRAY	"\\Gray{" 
#define GREEN	"\\OliveGreen{" 

int count;
int parens_level = 0;
int cons_init = 0;
int line_count = 0; 

/* != 0 <==> do line numbering */
int line_numbers = 0;

  void echo_text( char* t) {
    int i = 0;
    while ( t[i] != '\000') 
      switch ( t[i]) {
      case '{':
	printf( "%s", "\\begin{math}\\{\\end{math}");
	++i, ++parens_level;
	cons_init = 0;
	break;	
      case '}':
	printf( "%s", "\\begin{math}\\}\\end{math}");
	++i;
	if ( parens_level > 0)
	  --parens_level;
	break;	
      case '\134':
	printf( "%s", "\\begin{math}\\backslash\\end{math}");
	++i;
	break;	
      default:
	printf( "%c", t[i++]);
      }
  }
%}

%%

^{BOL} {
  BEGIN(MAY_KEY);
  echo_text( yytext); 
}

<MAY_KEY>{QUAL_FUNC_NAME}/{FUNC_DECL_FOLLOW} {
  if ( parens_level > 0 || cons_init != 0) {
    REJECT;
  }
  else {
    cons_init = 1; 
    BEGIN(0);
    printf( "%s{%s{", "\\textbf", "\\BROWN"); 
    echo_text( yytext); 
    printf( "%s", "}}");
  }
}

{COMMENT} {
  BEGIN(0);
  printf( "%s{%s{", "\\textsl", "\\RED"); 
  echo_text( yytext); 
  printf( "%s", "}}");
}
    
{STRING} {
  BEGIN(0);
  printf( "%s{", "\\TAN"); 
  echo_text( yytext); 
  printf( "%s", "}");
}
    
<MAY_KEY>{IMP_KEYWORD}/{FOLLOW_KEYWORD} {
  printf( "%s{%s{", "\\textbf", "\\VIOLET"); 
  echo_text( yytext); 
  printf( "%s", "}}");
}
      
<MAY_KEY>{ACCESS_SPEC}/{WHITE}*":" {
  printf( "%s{%s{", "\\textbf", "\\VIOLET"); 
  echo_text( yytext); 
  printf( "%s", "}}");
}
      
<MAY_KEY>{ACCESS_SPEC}/{WHITE}+{QUAL_ID} {
  BEGIN(MAY_BASE_CLASS_NAME);
  printf( "%s{%s{", "\\textbf", "\\BLUE"); 
  echo_text( yytext); 
  printf( "%s", "}}");
}
      
<MAY_KEY>{NAMESPACE}/{FULL_NAMESPACE_FOLLOW} {
  BEGIN(MAY_NAMESPACE);
  printf( "%s{%s{", "\\textbf", "\\BLUE"); 
  echo_text( yytext); 
  printf( "%s", "}}");
}

<MAY_NAMESPACE>{WHITE}+{ID}/{NAMESPACE_FOLLOW} {
  BEGIN(0);
  --parens_level;
  printf( "%s{%s{", "\\textbf", "\\BROWN"); 
  echo_text( yytext); 
  printf( "%s", "}}");
}

<MAY_KEY>{USING}/{FULL_USING_FOLLOW} {
  BEGIN(MAY_USING);
  printf( "%s{%s{", "\\textbf", "\\BLUE"); 
  echo_text( yytext); 
  printf( "%s", "}}");
}

<MAY_USING>{WHITE}+{NAMESPACE} {
  printf( "%s{%s{", "\\textbf", "\\BLUE"); 
  echo_text( yytext); 
  printf( "%s", "}}");
}

<MAY_USING>{WHITE}+{QUAL_ID}; {
  BEGIN(0);
  printf( "%s{%s{", "\\textbf", "\\BROWN"); 
  echo_text( yytext); 
  printf( "%s", "}}");
}

<MAY_KEY>{DECL_KEYWORD2}/{CLASS_DECL} {
  BEGIN(MAY_CLASS_DECL);
  printf( "%s{%s{", "\\textbf", "\\BLUE"); 
  echo_text( yytext); 
  printf( "%s", "}}");
}

<MAY_CLASS_DECL>{WHITE}+{ID}/{CLASS_DECL_FOLLOW1} {
  BEGIN(0);
  --parens_level;
  printf( "%s{%s{", "\\textbf", "\\BROWN"); 
  echo_text( yytext); 
  printf( "%s", "}}");
}

<MAY_CLASS_DECL>{WHITE}+{ID}/{CLASS_DECL_FOLLOW3} {
  BEGIN(0);
  cons_init = 1;
  --parens_level;
  printf( "%s{%s{", "\\textbf", "\\BROWN"); 
  echo_text( yytext); 
  printf( "%s", "}}");
}

<MAY_CLASS_DECL>{WHITE}+{ID}/{CLASS_DECL_FOLLOW2} {
  BEGIN(0);
  printf( "%s{%s{", "\\textbf", "\\BROWN"); 
  echo_text( yytext); 
  printf( "%s", "}}");
}

<MAY_KEY,MAY_TYPE_DECL>{DECL_KEYWORD1}/{FOLLOW_KEYWORD}	{
  BEGIN(MAY_TYPE_DECL);
  printf( "%s{%s{", "\\textbf", "\\BLUE"); 
  echo_text( yytext); 
  printf( "%s", "}}");
}

<MAY_TYPE_DECL>{QUAL_ID} {
  BEGIN(MAY_TYPE_NAME);
  echo_text( yytext); 
}

<MAY_BASE_CLASS_NAME>{WHITE}+{QUAL_ID} {
  BEGIN(0);
  printf( "%s{%s{", "\\textbf", "\\BROWN"); 
  echo_text( yytext); 
  printf( "%s", "}}");
}

<MAY_TYPE_NAME>{WHITE}+{ID}; {
  BEGIN(0);
  printf( "%s{%s{", "\\textbf", "\\BROWN"); 
  echo_text( yytext); 
  printf( "%s", "}}");
}

<MAY_KEY,MAY_TYPE_DECL>{TYPE_KEYWORD}/{FOLLOW_KEYWORD}	{
  printf( "%s{%s{", "\\textbf", "\\GBLUE"); 
  echo_text( yytext); 
  printf( "%s", "}}");
}

<MAY_KEY,MAY_TYPE_DECL>{MOD_KEYWORD}/{FOLLOW_KEYWORD}	{
  printf( "%s{%s{", "\\textbf", "\\LBLUE"); 
  echo_text( yytext); 
  printf( "%s", "}}");
}

<MAY_KEY,MAY_TYPE_DECL>{TEMPLATE_DECL}/{FOLLOW_KEYWORD}	{
  printf( "%s{%s{", "\\textbf", "\\MBLUE"); 
  echo_text( yytext); 
  printf( "%s", "}}");
}

<MAY_KEY>{OTHER_KEYWORD}/{FOLLOW_KEYWORD}	{
  printf( "%s{%s{", "\\textbf", "\\GRAY"); 
  echo_text( yytext); 
  printf( "%s", "}}");
}

{PREPROC1} {
  BEGIN(0);
  printf( "%s{%s{", "\\textbf", "\\MAGENTA");
  echo_text( yytext); 
  printf( "%s", "}}");
}

{PREPROC2} {
  BEGIN(0);
  printf( "%s{%s{", "\\textbf", "\\GREEN"); 
  echo_text( yytext); 
  printf( "%s", "}}");
}

\n/{WHITE_CR}*[^ \t\n] {
  BEGIN(MAY_KEY);
  echo_text( yytext); 
  if ( line_numbers) { printf( "%2d: ", ++line_count); }
}

\n {
  if ( line_numbers) { printf( "%2d: ", ++line_count); }
}

<MAY_KEY,MAY_TYPE_DECL>{PRECEDE_KEYWORD} {
  echo_text( yytext); 
}

<INITIAL>{PRECEDE_KEYWORD} {
  BEGIN(MAY_KEY);
  echo_text( yytext); 
}

. {
  BEGIN(0);
  echo_text( yytext); 
}

%%

int main( int argc, char* argv[])
{
    /* != 0 <==> print latex header */
    int header = 1;
    /* != 0 <==> use color */
    int color = 1;
    /* != 0 <==> colorize file */
    int work_on_file = 1;

    /* command line parsing */
    while ( --argc > 0) {
	if ( !strcmp( argv[argc], "-nh"))
	    header = 0;
	else if ( !strcmp( argv[argc], "-nc"))
	    color = 0;
	else if ( !strcmp( argv[argc], "-oh"))
	    work_on_file = 0;
	else if ( !strcmp( argv[argc], "-ln"))
	    line_numbers = 1;
	else
	    printf( "(%s) Unknown argument: \"%s\"\n",
		    argv[0],
		    argv[argc]);    
    }
    
    if ( header) {
	if ( work_on_file) {
	    printf( "%s\n", "%%% Automatically generated by cc2tex"); 
	    printf( "%s\n", "\\documentclass{article}");
	    printf( "%s\n", "\\usepackage{realA4,colordvi,alltt}"); 
	    printf( "%s\n", "\\pagestyle{empty}"); 
	    printf( "\n");
	}
	printf( "%s\n", "%% ----- cc2tex color definition -----");
	if ( color) {
	    printf( "%s\n", "%% for color print");
	    printf( "%s\n", "\\def\\RED#1{\\Mahogany{#1}}"); 
	    printf( "%s\n", "\\def\\BLUE#1{\\Orchid{#1}}"); 
	    printf( "%s\n", "\\def\\LBLUE#1{\\MidnightBlue{#1}}"); 
	    printf( "%s\n", "\\def\\MBLUE#1{\\Periwinkle{#1}}"); 
	    printf( "%s\n", "\\def\\GBLUE#1{\\CadetBlue{#1}}"); 
	    printf( "%s\n", "\\def\\BLACK#1{\\Black{#1}}"); 
	    printf( "%s\n", "\\def\\TAN#1{\\Tan{#1}}"); 
	    printf( "%s\n", "\\def\\VIOLET#1{\\Blue{#1}}"); 
	    printf( "%s\n", "\\def\\MAGENTA#1{\\Thistle{#1}}"); 
	    printf( "%s\n", "\\def\\BROWN#1{\\Plum{#1}}"); 
	    printf( "%s\n", "\\def\\GRAY#1{\\Gray{#1}}"); 
	    printf( "%s\n", "\\def\\GREEN#1{\\OliveGreen{#1}}"); 
	}
	else { 
	    printf( "%s\n", "%% for black&white print");
	    printf( "%s\n", "\\def\\RED#1{\\Mahogany{#1}}"); 
	    printf( "%s\n", "\\def\\BLUE#1{\\Orchid{#1}}"); 
	    printf( "%s\n", "\\def\\LBLUE#1{\\MidnightBlue{#1}}"); 
	    printf( "%s\n", "\\def\\MBLUE#1{\\Periwinkle{#1}}"); 
	    printf( "%s\n", "\\def\\GBLUE#1{\\CadetBlue{#1}}"); 
	    printf( "%s\n", "\\def\\BLACK#1{\\Black{#1}}"); 
	    printf( "%s\n", "\\def\\TAN#1{\\Tan{#1}}"); 
	    printf( "%s\n", "\\def\\VIOLET#1{\\Blue{#1}}"); 
	    printf( "%s\n", "\\def\\MAGENTA#1{\\Thistle{#1}}"); 
	    printf( "%s\n", "\\def\\BROWN#1{\\Plum{#1}}"); 
	    printf( "%s\n", "\\def\\GRAY#1{\\Gray{#1}}"); 
	    printf( "%s\n", "\\def\\GREEN#1{\\OliveGreen{#1}}"); 
	}
	printf( "\n");
	if ( work_on_file) {
	    printf( "%s\n", "\\begin{document}"); 
	    printf( "%s\n", "\\Large%SIZESET");
	    printf( "%s\n", "\\begin{alltt}"); 
	    if ( line_numbers) { printf( "%2d: ", ++line_count); }
	}
    }
    if ( work_on_file) {
	BEGIN(MAY_KEY);
	yylex();
	if ( header) {
	    printf( "%s\n", "\\end{alltt}"); 
	    printf( "%s\n", "\\end{document}"); 
	    printf( "%s\n", "%% EOF");
	}
    }
}

/* EOF */
