/* 
 *   Copyright (c) 2000 Michael Hoffmann (hoffmann@inf.ethz.ch)
 *   
 *   Permission is hereby granted, free of charge, to any person
 *   obtaining a copy of this software and associated documentation
 *   files (the "Software"), to deal in the Software without
 *   restriction, including without limitation the rights to use, copy,
 *   modify, merge, publish, distribute, sublicense, and/or sell copies
 *   of the Software, and to permit persons to whom the Software is
 *   furnished to do so, subject to the following conditions:
 *      
 *   The above copyright notice and this permission notice shall be
 *   included in all copies or substantial portions of the Software.
 *
 *   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 *   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 *   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 *   NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 *   BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 *   ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 *   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 *   SOFTWARE.
 *      
 *   $Revision: 1.6 $ 
 *   $Date: 2000/05/02 08:44:34 $
 *      
 *   convert Oberon/Modula source code to highlighted latex (using colordvi) 
 */

%option noyywrap
%s START_COND
%s MAY_PROCNAME
%s MAY_TYPE

COMMENT_BEGIN	[(][*]
COMMENT_END	[*][)]
STRING		\"[^"\n]*\"

WHITE		[ \t]
WHITE_CR	{WHITE}|\n
WHITE_COMMENT	({WHITE_CR}|{COMMENT})*
ALPHA		[a-zA-Z_]
ALNUM 		({ALPHA}|[[:digit:]])

D1_PAREN_EXPR 	"{"[^{}]*"}"
D2_PAREN_EXPR 	"{"({D1_PAREN_EXPR}|[^{}])*"}"
D3_PAREN_EXPR	"{"({D2_PAREN_EXPR}|[^{}])*"}"
D4_PAREN_EXPR	"{"({D3_PAREN_EXPR}|[^{}])*"}"
D5_PAREN_EXPR	"{"({D4_PAREN_EXPR}|[^{}])*"}"
D6_PAREN_EXPR	"{"({D5_PAREN_EXPR}|[^{}])*"}"
PAREN_EXPR 	{D6_PAREN_EXPR}

D1_RPAREN_EXPR 	"("[^()]*")"
D2_RPAREN_EXPR 	"("({D1_RPAREN_EXPR}|[^()])*")"
D3_RPAREN_EXPR	"("({D2_RPAREN_EXPR}|[^()])*")"
D4_RPAREN_EXPR	"("({D3_RPAREN_EXPR}|[^()])*")"
D5_RPAREN_EXPR	"("({D4_RPAREN_EXPR}|[^()])*")"
D6_RPAREN_EXPR	"("({D5_RPAREN_EXPR}|[^()])*")"
RPAREN_EXPR 	{D6_RPAREN_EXPR}

ID		{ALPHA}{ALNUM}*
QUAL_ID		({ID}[.])?{ID}
ID_DEF		{ID}({WHITE}*[*-])?

    /* these have to follow a keyword */
FOLLOW_KEYWORD	[ (\n\t;)]

    /* important keywords */
IMP_KEYWORD	(CASE|OF|DO|ELSE|ELSIF|FOR|EXIT|IF|LOOP|REPEAT|RETURN|THEN|WHILE|WITH)

    /* keywords opening declarations */
DECL_KEYWORD	(IMPORT|BEGIN|END)

    /* types */
TYPE_KEYWORD	(POINTER{WHITE}+TO{WHITE}+{QUAL_ID})

    /* predeclared procedures */
PREDEF_KEYWORD	(ASSERT|COPY|DEC|EXCL|HALT|INC|INCL|NEW|ABS|ASH|CAP|CHR|ENTIER|LEN|LONG|MAX|MIN|ODD|ORD|SHORT|SIZE)

ID_LIST			{ID}({WHITE}*[,]{WHITE}*{ID})*
FUNC_ARG 		(VAR{WHITE}+)?{ID_LIST}{WHITE}*[:]{WHITE}*{QUAL_ID}
FUNC_ARGLIST 		[(]{WHITE}*{FUNC_ARG}({WHITE}*[;]{WHITE_CR}*{FUNC_ARG})*{WHITE}*[)]({WHITE_CR}*[:]{QUAL_ID})*
RECEIVER		[(]{WHITE}*((VAR){WHITE}*)?{ID}{WHITE}*[:]{WHITE}*{ID}{WHITE}*[)]
FUNCTION		{ID_DEF}
FUNCTION_FOLLOW		({WHITE}*{FUNC_ARGLIST})?

%{
#define RED	"\\Mahogany{" 
#define BLUE	"\\Orchid{" 
#define LBLUE	"\\MidnightBlue{" 
#define MBLUE	"\\Periwinkle{" 
#define GBLUE	"\\CadetBlue{" 
#define BLACK	"\\Black{" 
#define TAN	"\\Tan{" 
#define VIOLET	"\\Blue{" 
#define MAGENTA	"\\Thistle{" 
#define BROWN	"\\Plum{" 
#define GRAY	"\\Gray{" 
#define GREEN	"\\OliveGreen{" 

int count;
int line_count = 0; 
int comment = 0;
int typedecl = 0;
int array = 0;

  void echo_text( char* t) {
    int i = 0;
    while ( t[i] != '\000') 
      switch ( t[i]) {
      case '{':
	printf( "%s", "\\begin{math}\\{\\end{math}");
	++i;
	break;	
      case '}':
	printf( "%s", "\\begin{math}\\}\\end{math}");
	++i;
	break;	
      case '\134':
	printf( "%s", "\\begin{math}\\backslash\\end{math}");
	++i;
	break;	
      default:
	printf( "%c", t[i++]);
      }
  }

  void colorize(char* c) {
    printf( "%s{%s{", "\\textbf", c); 
    echo_text( yytext); 
    printf( "%s", "}}");
  }

%}

%%

{WHITE}+ {
  echo_text( yytext); 
}

{COMMENT_BEGIN} {
  BEGIN(START_COND);
  comment = 1;
  printf( "%s{%s{", "\\textsl", "\\RED"); 
  echo_text( yytext); 
}
    
{COMMENT_END} {
  BEGIN(START_COND);
  comment = 0;
  echo_text( yytext); 
  printf( "%s", "}}");
}
    
([^*\n]|[*][^)\n])+ {
  if (comment == 0) { REJECT; }
  echo_text( yytext); 
}

[=]{WHITE}*/{QUAL_ID} {
  if (typedecl == 0) { 
    REJECT; 
  } else {
    BEGIN(MAY_TYPE);
    echo_text( yytext); 
  }
}

ARRAY {
  BEGIN(START_COND);
  ++array;
  colorize("\\GBLUE");
}

{ALNUM}+({WHITE}*[,]{WHITE}*{ALNUM}+)* {
  if (array != 1) { REJECT; }
  ++array;
  echo_text( yytext); 
}

OF/{FOLLOW_KEYWORD} {
  if (array != 2) { REJECT; }
  array = 0;
  BEGIN(MAY_TYPE);
  colorize("\\GBLUE");
}

PROCEDURE {
  BEGIN(MAY_PROCNAME);
  typedecl = 0;
  colorize("\\BLUE");
}

MODULE {
  BEGIN(MAY_PROCNAME);
  colorize("\\BLUE");
}

(TYPE) {
  BEGIN(START_COND);
  typedecl = 1;
  colorize("\\BLUE");
}

RECORD {
  BEGIN(MAY_TYPE);
  colorize("\\GBLUE");
}

<MAY_TYPE>[(]{WHITE}*/{QUAL_ID}{WHITE}*[)] {
  echo_text( yytext); 
}

{ID_DEF}/{WHITE}*[=] {
  if (typedecl == 0) { 
    REJECT; 
  } else 
    colorize("\\GBLUE");
}

(CONST) {
  BEGIN(START_COND);
  typedecl = 0;
  colorize("\\BLUE");
}

(VAR) {
  BEGIN(START_COND);
  typedecl = 0;
  colorize("\\GBLUE");
}

<MAY_PROCNAME>{RECEIVER}{WHITE}* {
  echo_text( yytext); 
}

<MAY_PROCNAME>{FUNCTION}/{FUNCTION_FOLLOW} {
  BEGIN(START_COND);
  colorize("\\BROWN");
}

END/{WHITE}+{ID_DEF} {
  BEGIN(MAY_PROCNAME);
  colorize("\\BLUE");
}

{IMP_KEYWORD}/{FOLLOW_KEYWORD} {
  BEGIN(START_COND);
  colorize("\\VIOLET");
}
      
{DECL_KEYWORD}/{FOLLOW_KEYWORD} {
  BEGIN(START_COND);
  colorize("\\BLUE");
}

{TYPE_KEYWORD}/{FOLLOW_KEYWORD}	{
  BEGIN(START_COND);
  colorize("\\GBLUE");
}

{PREDEF_KEYWORD}/{WHITE}*[(] {
  BEGIN(START_COND);
  colorize("\\GRAY");
}

{STRING} {
  BEGIN(START_COND);
  printf( "%s{", "\\TAN"); 
  echo_text( yytext); 
  printf( "%s", "}");
}
    
<START_COND>[:]/{WHITE}*{QUAL_ID} {
  BEGIN(MAY_TYPE);
  echo_text( yytext); 
}

<MAY_TYPE>{QUAL_ID} {
  BEGIN(START_COND);
  colorize("\\GBLUE");
}

\n {
  BEGIN(START_COND);
  if (comment == 1)
    printf( "%s", "}}");
  echo_text( yytext); 
  printf( "%3d: ", ++line_count); 
  if (comment == 1)
    printf( "%s{%s{", "\\textsl", "\\RED"); 
}

. {
  BEGIN(START_COND);
  echo_text( yytext); 
}

%%

int main( int argc, char* argv[])
{
    /* != 0 <==> print latex header */
    int header = 1;
    /* != 0 <==> use color */
    int color = 1;
    /* != 0 <==> colorize file */
    int work_on_file = 1;

    /* command line parsing */
    while ( --argc > 0) {
	if ( !strcmp( argv[argc], "-nh"))
	    header = 0;
	else if ( !strcmp( argv[argc], "-nc"))
	    color = 0;
	else if ( !strcmp( argv[argc], "-oh"))
	    work_on_file = 0;
	else
	    printf( "(%s) Unknown argument: \"%s\"\n",
		    argv[0],
		    argv[argc]);    
    }
    
    if ( header) {
	if ( work_on_file) {
	    printf( "%s\n", "%%% Automatically generated by mod2tex"); 
	    printf( "%s\n", "\\documentclass{article}");
	    printf( "%s\n", "\\usepackage{realA4,colordvi,alltt}"); 
	    printf( "%s\n", "\\pagestyle{empty}"); 
	    printf( "\n");
	}
	printf( "%s\n", "%% ----- mod2tex color definition -----");
	if ( color) {
	    printf( "%s\n", "%% for color print");
	    printf( "%s\n", "\\def\\RED#1{\\Mahogany{#1}}"); 
	    printf( "%s\n", "\\def\\BLUE#1{\\Orchid{#1}}"); 
	    printf( "%s\n", "\\def\\LBLUE#1{\\MidnightBlue{#1}}"); 
	    printf( "%s\n", "\\def\\MBLUE#1{\\Periwinkle{#1}}"); 
	    printf( "%s\n", "\\def\\GBLUE#1{\\CadetBlue{#1}}"); 
	    printf( "%s\n", "\\def\\BLACK#1{\\Black{#1}}"); 
	    printf( "%s\n", "\\def\\TAN#1{\\Tan{#1}}"); 
	    printf( "%s\n", "\\def\\VIOLET#1{\\Blue{#1}}"); 
	    printf( "%s\n", "\\def\\MAGENTA#1{\\Thistle{#1}}"); 
	    printf( "%s\n", "\\def\\BROWN#1{\\Plum{#1}}"); 
	    printf( "%s\n", "\\def\\GRAY#1{\\Gray{#1}}"); 
	    printf( "%s\n", "\\def\\GREEN#1{\\OliveGreen{#1}}"); 
	}
	else { 
	    printf( "%s\n", "%% for black&white print");
	    printf( "%s\n", "\\def\\RED#1{\\Mahogany{#1}}"); 
	    printf( "%s\n", "\\def\\BLUE#1{\\Orchid{#1}}"); 
	    printf( "%s\n", "\\def\\LBLUE#1{\\MidnightBlue{#1}}"); 
	    printf( "%s\n", "\\def\\MBLUE#1{\\Periwinkle{#1}}"); 
	    printf( "%s\n", "\\def\\GBLUE#1{\\CadetBlue{#1}}"); 
	    printf( "%s\n", "\\def\\BLACK#1{\\Black{#1}}"); 
	    printf( "%s\n", "\\def\\TAN#1{\\Tan{#1}}"); 
	    printf( "%s\n", "\\def\\VIOLET#1{\\Blue{#1}}"); 
	    printf( "%s\n", "\\def\\MAGENTA#1{\\Thistle{#1}}"); 
	    printf( "%s\n", "\\def\\BROWN#1{\\Plum{#1}}"); 
	    printf( "%s\n", "\\def\\GRAY#1{\\Gray{#1}}"); 
	    printf( "%s\n", "\\def\\GREEN#1{\\OliveGreen{#1}}"); 
	}
	printf( "\n");
	if ( work_on_file) {
	    printf( "%s\n", "\\begin{document}"); 
	    printf( "%s\n", "\\Large%SIZESET");
	    printf( "%s\n", "\\begin{alltt}"); 
	    printf( "%3d: ", ++line_count); 
	}
    }
    if ( work_on_file) {
        BEGIN(START_COND);
	yylex();
	if ( header) {
	    printf( "%s\n", "\\end{alltt}"); 
	    printf( "%s\n", "\\end{document}"); 
	    printf( "%s\n", "%% EOF");
	}
    }
}

/* EOF */
